unit basep4;
// ----------------------------------------------------------------------------
interface
// ----------------------------------------------------------------------------
const
  NbJoueurs         = 2;
  NbCasesParColonne = 6; // Plateau vertical de 6x7 cases
  NbCasesParLigne   = 7; // Plateau vertical de 6x7 cases
  NbMaxTours        = NbCasesParColonne*NbCasesParLigne; // Max = nombre de cases !
// ----------------------------------------------------------------------------
type
  TDirection = (dIndefinie,
                dHaut, dHautDroite, dDroite, dBasDroite,
                dBas,  dBasGauche,  dGauche, dHautGauche);
  TLigne     = 1..NbCasesParColonne;
  TColonne   = 1..NbCasesParLigne;
  TTour      = 0..NbMaxTours;
  TCase      = (cPersonne, cJoueur1, cJoueur2); // 0=vide ; 1=Joueur qui commence ; 2=l'autre joueur
  TPartie    = array [TTour, TColonne, TLigne] of TCase;
  TDestinee  = array [TTour, TColonne] of TCase;
// ----------------------------------------------------------------------------
  TPuissance4 = class
  private
    VersionFichier : Integer; // En fait une constante mais FileRead/Write n'aime pas les constantes
    function CoordonneesCorrectes(const x, y : Integer) : Boolean;
  public
    Tour, // Tour courant
    DernierTour : TTour; // Dernier tour atteint tenant compte des actions dfaire
    NbMemesPionsAvant,  // positionnement relatif des extrmits de l'alignement
    NbMemesPionsApres : Integer;
    dAlignement : TDirection;
    Vainqueur : TCase;
    p : TPartie; // Stockage des pions jous par tour, colonne & ligne
    constructor Create;
    procedure PrepareTourSuivant;
    function JoueurCourant : TCase;
    function QuatrePionsAlignes(const x : TColonne; const y : TLigne) : TDirection;
    function OuvrePartie(const stCheminNomFichier : String) : Boolean;
    function EnregistrePartie(const stCheminNomFichier : String) : Boolean;
  end{class TPuissance4};
// ----------------------------------------------------------------------------
const
  dx     : array [TDirection] of Integer = (0, 0, 1, 1, 1, 0, -1, -1, -1);
  dy     : array [TDirection] of Integer = (0, -1, -1, 0, 1, 1, 1, 0, -1);
  InvDir : array [TDirection] of TDirection = (dIndefinie,
                                               dBas,  dBasGauche,  dGauche, dHautGauche,
                                               dHaut, dHautDroite, dDroite, dBasDroite);
// ----------------------------------------------------------------------------
implementation
// ----------------------------------------------------------------------------
uses
  Math, Windows, SysUtils;
// ----------------------------------------------------------------------------
constructor TPuissance4.Create;
begin
VersionFichier:=1; // Version fichier permettant de grer les volutions
Tour:=0; DernierTour:=0;
dAlignement:=dIndefinie;
NbMemesPionsAvant:=0;
NbMemesPionsApres:=0;
Vainqueur:=cPersonne;
end;
// ----------------------------------------------------------------------------
procedure TPuissance4.PrepareTourSuivant;
var i : TColonne;
    j : TLigne;
begin
Inc(Tour);
DernierTour:=Max(Tour, DernierTour);
for i:=Low(TColonne) to High(TColonne) do
  for j:=Low(TLigne) to High(TLigne) do
    p[Tour,i,j]:=p[Tour-1,i,j];
end;
// ----------------------------------------------------------------------------
function TPuissance4.JoueurCourant : TCase;
begin
Result:=TCase(2-Tour mod 2);
end;
// ----------------------------------------------------------------------------
function TPuissance4.CoordonneesCorrectes(const x, y : Integer) : Boolean;
begin
Result:=(x>=Low(TColonne)) and (x<=High(TColonne)) and
        (y>=Low(TLigne))   and (y<=High(TLigne));
end;
// ----------------------------------------------------------------------------
function TPuissance4.QuatrePionsAlignes(const x : TColonne; const y : TLigne) : TDirection;
var JoueurPion : TCase;
begin // On part du dernier pion lch et on regarde de chaque ct la mme couleur pour chaque directions (en fait 4 modulo les sens)
JoueurPion:=p[Tour,x,y];
dAlignement:=dIndefinie;
NbMemesPionsAvant:=0;
NbMemesPionsApres:=0;
while (dAlignement<>dBasDroite) and
      (NbMemesPionsAvant+NbMemesPionsApres+1<4) do
  begin
  Inc(dAlignement);
  NbMemesPionsAvant:=0;
  NbMemesPionsApres:=0;
  while CoordonneesCorrectes(x+(NbMemesPionsApres+1)*dx[dAlignement],
                             y+(NbMemesPionsApres+1)*dy[dAlignement]) and
        (p[Tour,
           x+(NbMemesPionsApres+1)*dx[dAlignement],
           y+(NbMemesPionsApres+1)*dy[dAlignement]]=JoueurPion) do
   Inc(NbMemesPionsApres);
  while CoordonneesCorrectes(x+(NbMemesPionsAvant+1)*dx[InvDir[dAlignement]],
                             y+(NbMemesPionsAvant+1)*dy[InvDir[dAlignement]]) and
        (p[Tour,
           x+(NbMemesPionsAvant+1)*dx[InvDir[dAlignement]],
           y+(NbMemesPionsAvant+1)*dy[InvDir[dAlignement]]]=JoueurPion) do
   Inc(NbMemesPionsAvant);
   end;
if NbMemesPionsAvant+NbMemesPionsApres+1>=4 then
  begin
  Result:=dAlignement;
  Vainqueur:=JoueurPion;
  end
else
  Result:=dIndefinie;
end{QuatrePionsAlignes};
// ----------------------------------------------------------------------------
function TPuissance4.OuvrePartie(const stCheminNomFichier : String) : Boolean;
var f : Integer;
    t : TTour;
    x : TColonne;
    y : TLigne;
begin
Result:=False; // Pessimiste !
f:=FileOpen(stCheminNomFichier, fmOpenRead);
try
  Result:=(FileRead(f, VersionFichier, SizeOf(VersionFichier))>0) and
          (FileRead(f, Tour, SizeOf(Tour))>0) and
          (FileRead(f, DernierTour, SizeOf(DernierTour))>0) and
          (FileRead(f, NbMemesPionsAvant, SizeOf(NbMemesPionsAvant))>0) and
          (FileRead(f, NbMemesPionsApres, SizeOf(NbMemesPionsApres))>0) and
          (FileRead(f, dAlignement, SizeOf(dAlignement))>0) and
          (FileRead(f, Vainqueur, SizeOf(Vainqueur))>0);
  if Result then
    for t:=Low(TTour) to High(TTour) do
      for x:=Low(TColonne) to High(TColonne) do
        for y:=Low(TLigne) to High(TLigne) do
          if Result then Result:=(FileRead(f, p[t, x, y], SizeOf(TCase))>0);
finally
  FileClose(f);
end{try};
end;
// ----------------------------------------------------------------------------
function TPuissance4.EnregistrePartie(const stCheminNomFichier : String) : Boolean;
var f : Integer;
    t : TTour;
    x : TColonne;
    y : TLigne;
begin
Result:=False; // Pessimiste !
if FileExists(stCheminNomFichier) then
  DeleteFile(stCheminNomFichier);
f:=FileCreate(stCheminNomFichier, fmOpenWrite);
try
  Result:=(FileWrite(f, VersionFichier, SizeOf(VersionFichier))>0) and
          (FileWrite(f, Tour, SizeOf(Tour))>0) and
          (FileWrite(f, DernierTour, SizeOf(DernierTour))>0) and
          (FileWrite(f, NbMemesPionsAvant, SizeOf(NbMemesPionsAvant))>0) and
          (FileWrite(f, NbMemesPionsApres, SizeOf(NbMemesPionsApres))>0) and
          (FileWrite(f, dAlignement, SizeOf(dAlignement))>0) and
          (FileWrite(f, Vainqueur, SizeOf(Vainqueur))>0);
  if Result then
    for t:=Low(TTour) to High(TTour) do
      for x:=Low(TColonne) to High(TColonne) do
        for y:=Low(TLigne) to High(TLigne) do
          if Result then Result:=(FileWrite(f, p[t, x, y], SizeOf(TCase))>0);
finally
  FileClose(f);
end{try};
end;
// ----------------------------------------------------------------------------
end.
